import numpy as np
import numpy.typing as npt
from dataclasses import dataclass, field
from abc import ABC
from ..SignalProcessing.periodic_functions import PeriodicFunction, HarmonicCoefficients, fourier_series, CosFunction
from typing import Type
    
@dataclass(frozen=True)
class Component(ABC):
    nodes : tuple[str]
    type : str
    id : str
    is_active: bool = field(default=False, init=False)
    w: float = field(default=0, init=False)

@dataclass(frozen=True)
class TwoPoleComponent(Component):
    nodes : tuple[str, str]

@dataclass(frozen=True)
class Resistor(TwoPoleComponent):
    R : float
    type : str = field(default='resistor', init=False)

@dataclass(frozen=True)
class Conductance(TwoPoleComponent):
    G : float
    type : str = field(default='conductance', init=False)

@dataclass(frozen=True)
class Capacitor(TwoPoleComponent):
    C : float
    type : str = field(default='capacitor', init=False)

@dataclass(frozen=True)
class Inductance(TwoPoleComponent):
    L : float
    type : str = field(default='inductance', init=False)

@dataclass(frozen=True)
class Source(TwoPoleComponent):
    is_active : bool = field(default=True, init=False)
    w : float = field(default=0.0)
    phi : float = field(default=0.0)

@dataclass(frozen=True)
class CurrentSource(Source):
    I : float = field(default=0.0)
    type : str = field(default='current_source', init=False)

@dataclass(frozen=True)
class VoltageSource(Source):
    V : float = field(default=0.0)
    type : str = field(default='voltage_source', init=False)

@dataclass(frozen=True)
class LinearCurrentSource(CurrentSource):
    G : float = field(default=0.0)
    type : str = field(default='linear_current_source', init=False)

@dataclass(frozen=True)
class LinearVoltageSource(VoltageSource):
    R : float = field(default=0.0)
    type : str = field(default='linear_voltage_source', init=False)

@dataclass(frozen=True)
class PeriodicVoltageSource(VoltageSource):
    wavetype : Type[PeriodicFunction] = field(default=CosFunction)

    @property
    def time_properties(self) -> PeriodicFunction:
        return self.wavetype(period=2*np.pi/self.w, amplitude=self.V, phase=self.phi)

    @property
    def frequency_properties(self) -> HarmonicCoefficients:
        return fourier_series(self.time_properties)

    def frequency_components(self, w_max: float) -> npt.NDArray[np.double]:
        n_max = np.floor(w_max/self.w)
        return np.array([self.w*n for n in np.arange(n_max)])

@dataclass(frozen=True)
class PeriodicCurrentSource(CurrentSource):
    wavetype : Type[PeriodicFunction] = field(default=CosFunction)

    @property
    def time_properties(self) -> PeriodicFunction:
        return self.wavetype(period=2*np.pi/self.w, amplitude=self.I, phase=self.phi)

    @property
    def frequency_properties(self) -> HarmonicCoefficients:
        return fourier_series(self.time_properties)

    def frequency_components(self, w_max: float) -> npt.NDArray[np.double]:
        n_max = np.floor(w_max/self.w)
        return np.array([self.w*n for n in np.arange(n_max)])

@dataclass(frozen=True)
class Ground(Component):
    id : str = field(default='gnd')
    type : str = field(default='ground', init=False)
    nodes : tuple[str]


import numpy as np
from ..Network import elements as elm
from ..Network import network as ntw
from typing import Callable, Type, TypeVar
from . import components as cmp

CircuitComponent = TypeVar("CircuitComponent", bound=cmp.Component)
CircuitComponentTranslator = Callable[[CircuitComponent, float, float], ntw.Branch]

def resistor(resistor: cmp.Resistor, *_) -> ntw.Branch:
    return ntw.Branch(resistor.nodes[0], resistor.nodes[1], elm.resistor(resistor.id, resistor.R))

def capacitor(capacitor: cmp.Capacitor, w: float = 0, *_) -> ntw.Branch:
    return ntw.Branch(
        capacitor.nodes[0],
        capacitor.nodes[1],
        elm.admittance(capacitor.id, elm.admittance_value(B=w*capacitor.C)))

def inductance(inductance: cmp.Inductance, w: float = 0, *_) -> ntw.Branch:
    return ntw.Branch(
        inductance.nodes[0],
        inductance.nodes[1],
        elm.impedance(inductance.id, elm.impedance_value(X=w*inductance.L))
    )

def current_source(current_source: cmp.CurrentSource, w: float = 0, w_resolution: float = 1e-3) -> ntw.Branch:
    element = elm.current_source(current_source.id, elm.complex_value(current_source.I, 0))
    if np.abs(w-current_source.w) > w_resolution:
        element = elm.open_circuit(current_source.id)
    return ntw.Branch(
        current_source.nodes[0],
        current_source.nodes[1],
        element
    )

def voltage_source(voltage_source: cmp.VoltageSource, w: float = 0, w_resolution: float = 1e-3) -> ntw.Branch:
    element = elm.voltage_source(voltage_source.id, elm.complex_value(voltage_source.V, voltage_source.phi))
    if np.abs(w-voltage_source.w) > w_resolution:
        element = elm.short_cicruit(voltage_source.id)
    return ntw.Branch(
        voltage_source.nodes[0],
        voltage_source.nodes[1],
        element)

def periodic_voltage_source(source: cmp.PeriodicVoltageSource, w: float = 0, w_resolution: float = 1e-3) -> ntw.Branch:
    n = np.round(w/source.w)
    delta_n = np.abs(w/source.w - n)
    if delta_n > w_resolution/source.w:
        return ntw.Branch(
            source.nodes[0],
            source.nodes[1],
            elm.short_cicruit(source.id))
    single_frequency_source = cmp.VoltageSource(
        nodes=source.nodes,
        id=source.id,
        w=w,
        phi=source.frequency_properties.phase(n),
        V=source.frequency_properties.amplitude(n)
    )
    return voltage_source(single_frequency_source, w, w_resolution)

def periodic_current_source(source: cmp.PeriodicCurrentSource, w: float = 0, w_resolution: float = 1e-3) -> ntw.Branch:
    n = np.round(w/source.w)
    delta_n = np.abs(w/source.w - n)
    if delta_n > w_resolution/source.w:
        return ntw.Branch(
            source.nodes[0],
            source.nodes[1],
            elm.open_circuit(source.id))
    single_frequency_source = cmp.CurrentSource(
        nodes=source.nodes,
        id=source.id,
        w=w,
        phi=source.frequency_properties.phase(n),
        I=source.frequency_properties.amplitude(n)
    )
    return current_source(single_frequency_source, w, w_resolution)

def linear_current_source(current_source: cmp.LinearCurrentSource, w: float = 0, w_resolution: float = 1e-3) -> ntw.Branch:
    element = elm.linear_current_source(current_source.id, elm.complex_value(current_source.I, 0), elm.complex_value(current_source.G, 0))
    if np.abs(w-current_source.w) > w_resolution:
        element = elm.open_circuit(current_source.id)
    return ntw.Branch(
        current_source.nodes[0],
        current_source.nodes[1],
        element)

def linear_voltage_source(voltage_source: cmp.LinearVoltageSource, w: float = 0, w_resolution: float = 1e-3) -> ntw.Branch:
    element = elm.linear_voltage_source(voltage_source.id, elm.complex_value(voltage_source.V, 0), elm.complex_value(voltage_source.R, 0))
    if np.abs(w-voltage_source.w) > w_resolution:
        element = elm.short_cicruit(voltage_source.id)
    return ntw.Branch(
        voltage_source.nodes[0],
        voltage_source.nodes[1],
        element
    )

transformers : dict[Type[cmp.Component], CircuitComponentTranslator] = {
    cmp.Resistor : resistor,
    cmp.Capacitor : capacitor,
    cmp.Inductance : inductance,
    cmp.CurrentSource : current_source,
    cmp.VoltageSource : voltage_source,
    cmp.LinearCurrentSource : linear_current_source,
    cmp.LinearVoltageSource : linear_voltage_source,
    cmp.PeriodicVoltageSource : periodic_voltage_source,
    cmp.PeriodicCurrentSource : periodic_current_source,
}


from .circuit import Circuit, transform, frequency_components
from ..SignalProcessing.types import TimeDomainFunction, FrequencyDomainFunction
from ..Network.NodalAnalysis import nodal_analysis_solver
from ..Network.solution import NetworkSolver
from typing import Any
from dataclasses import dataclass, field
from abc import ABC, abstractmethod
import numpy as np

@dataclass
class CircuitSolution(ABC):
    circuit: Circuit = field(default_factory=lambda : Circuit([]))
    solver: NetworkSolver = field(default=nodal_analysis_solver)

    @abstractmethod
    def get_voltage(self, id: str) -> Any:
        ...

    @abstractmethod
    def get_current(self, id: str) -> Any:
        ...

    @abstractmethod
    def get_power(self, id: str) -> Any:
        ...

@dataclass
class DCSolution(CircuitSolution):
    def __post_init__(self):
        network = transform(self.circuit, w=[0])[0]
        self._solution = self.solver(network)

    def get_voltage(self, component_id: str) -> float:
        return self._solution.get_voltage(component_id).real

    def get_current(self, component_id: str) -> float:
        return self._solution.get_current(component_id).real

    def get_power(self, component_id: str) -> float:
        return self.get_voltage(component_id)*self.get_current(component_id)

@dataclass
class ComplexSolution(CircuitSolution):
    w: float = 0

    def __post_init__(self):
        network = transform(self.circuit, w=[self.w])[0]
        self._solution = self.solver(network)

    def get_voltage(self, component_id: str) -> complex:
        return self._solution.get_voltage(component_id)

    def get_current(self, component_id: str) -> complex:
        return self._solution.get_current(component_id)

    def get_power(self, component_id: str) -> complex:
        return self._solution.get_voltage(component_id)*np.conj(self._solution.get_current(component_id))

@dataclass
class TimeDomainSolution(CircuitSolution):
    w_max: float = field(default=0)
    w: list[float] = field(default_factory=list, init=False)

    def __post_init__(self):
        self.w = frequency_components(self.circuit, self.w_max)
        networks = transform(self.circuit, w=self.w)
        self._solutions = [self.solver(network) for network in networks]

    def get_voltage(self, component_id: str) -> TimeDomainFunction:
        voltages = [solution.get_voltage(component_id) for solution in self._solutions]
        return np.vectorize(lambda t: np.array(np.sum([np.abs(V)*np.cos(w*t+np.angle(V)) for V, w in zip(voltages, self.w)])))

    def get_current(self, component_id: str) -> TimeDomainFunction:
        currents = [solution.get_current(component_id) for solution in self._solutions]
        return np.vectorize(lambda t: np.array(np.sum([np.abs(V)*np.cos(w*t+np.angle(V)) for V, w in zip(currents, self.w)])))

    def get_power(self, component_id: str) -> TimeDomainFunction:
        voltage = self.get_voltage(component_id)
        current = self.get_current(component_id)
        return lambda t: np.array(voltage(t))*np.array(current(t))

@dataclass
class FrequencyDomainSolution(CircuitSolution):
    w_max: float = field(default=0)

    def __post_init__(self):
        self.w = frequency_components(self.circuit, self.w_max)
        networks = transform(self.circuit, w=self.w)
        self._solutions = [self.solver(network) for network in networks]

    def get_voltage(self, component_id: str) -> FrequencyDomainFunction:
        voltages = np.array([solution.get_voltage(component_id) for solution in self._solutions])
        return np.array(self.w), voltages

    def get_current(self, component_id: str) -> FrequencyDomainFunction:
        currents = np.array([solution.get_current(component_id) for solution in self._solutions])
        return np.array(self.w), currents

    def get_power(self, component_id: str) -> FrequencyDomainFunction:
        w, voltage = self.get_voltage(component_id)
        _, current = self.get_current(component_id)
        return w, voltage*np.conj(current)


from . import Elements as elm
from ..Circuit import components as cct_cmp
from ..SignalProcessing import periodic_functions
from .SchemdrawTranslatorTypes import ElementTranslatorMap
from math import pi, inf


def resistor_translator(element: elm.Resistor, nodes: tuple[str, str]) -> cct_cmp.Resistor:
    return cct_cmp.Resistor(nodes=nodes, id=element.name, R=element.R)

def conductance_translator(element: elm.Conductance, nodes: tuple[str, str]) -> cct_cmp.Conductance:
    return cct_cmp.Conductance(nodes=nodes, id=element.name, G=element.G)

def capacitor_translator(element: elm.Capacitor, nodes: tuple[str, str]) -> cct_cmp.Capacitor:
    return cct_cmp.Capacitor(nodes=nodes, id=element.name, C=element.C)

def inductance_translator(element: elm.Inductance, nodes: tuple[str, str]) -> cct_cmp.Inductance:
    return cct_cmp.Inductance(nodes=nodes, id=element.name, L=element.L)

def ground_translator(element: elm.Ground, nodes: tuple[str]) -> cct_cmp.Ground:
    return cct_cmp.Ground(nodes=nodes, id=element.name)

def dc_voltage_source_translator(element: elm.VoltageSource, nodes: tuple[str, str]) -> cct_cmp.VoltageSource:
    return cct_cmp.VoltageSource(
        nodes=nodes if not element.is_reverse else nodes[::-1],
        id=element.name,
        V=element.V.real if not element.is_reverse else -element.V.real,
        w=0,
        phi=0
    )

def dc_current_source_translator(element: elm.CurrentSource, nodes: tuple[str, str]) -> cct_cmp.CurrentSource:
    return cct_cmp.CurrentSource(
        nodes=nodes if not element.is_reverse else nodes[::-1],
        id=element.name,
        I=element.I.real if not element.is_reverse else -element.I.real,
        w=0,
        phi=0
    )

def ac_voltage_source_translator(element: elm.ACVoltageSource, nodes: tuple[str, str]) -> cct_cmp.VoltageSource:
    return cct_cmp.VoltageSource(
        nodes=nodes if not element.is_reverse else nodes[::-1],
        id=element.name,
        V=element.V if not element.is_reverse else -element.V,
        w=element.w,
        phi=element.phi*pi/180 if element.deg else element.phi
    )

def ac_current_source_translator(element: elm.ACCurrentSource, nodes: tuple[str, str]) -> cct_cmp.CurrentSource:
    return cct_cmp.CurrentSource(
        nodes=nodes if not element.is_reverse else nodes[::-1],
        id=element.name,
        I=element.I if not element.is_reverse else -element.I,
        w=element.w,
        phi=element.phi*pi/180 if element.deg else element.phi
    )

def rect_voltage_source_translator(element: elm.RectVoltageSource, nodes: tuple[str, str]) -> cct_cmp.VoltageSource:
    return cct_cmp.PeriodicVoltageSource(
        nodes=nodes[::-1] if not element.is_reverse else nodes,
        id=element.name,
        V=element.V if not element.is_reverse else -element.V,
        w=element.w,
        phi=element.phi*pi/180 if element.deg else element.phi,
        wavetype=periodic_functions.RectFunction
    )

def rect_current_source_translator(element: elm.RectCurrentSource, nodes: tuple[str, str]) -> cct_cmp.CurrentSource:
    return cct_cmp.PeriodicCurrentSource(
        nodes=nodes if not element.is_reverse else nodes[::-1],
        id=element.name,
        I=element.I if not element.is_reverse else -element.I,
        w=element.w,
        phi=element.phi*pi/180 if element.deg else element.phi,
        wavetype=periodic_functions.RectFunction
    )

def linear_current_source_translator(element: elm.RealCurrentSource, nodes: tuple[str, str]) -> cct_cmp.LinearCurrentSource:
    return cct_cmp.LinearCurrentSource(
        nodes=nodes if not element.is_reverse else nodes[::-1],
        id=element.name,
        I=element.I.real if not element.is_reverse else -element.I.real,
        G=element.G)

def linear_voltage_source_translator(element: elm.RealVoltageSource, nodes: tuple[str, str]) -> cct_cmp.LinearVoltageSource:
    return cct_cmp.LinearVoltageSource(
        nodes=nodes if not element.is_reverse else nodes[::-1],
        id=element.name,
        V=element.V.real if not element.is_reverse else -element.V.real,
        R=element.R
    )

def switch_translator(element: elm.Switch, nodes: tuple[str, str]) -> cct_cmp.Resistor | None:
    if element.state == element.state.OPEN:
        return cct_cmp.Resistor(nodes=nodes, id=element.name, R=inf)
    return cct_cmp.Resistor(nodes=nodes, id=element.name, R=1e-12)

def none_translator(*_) -> None:
    return None

circuit_translator_map : ElementTranslatorMap = {
    elm.Resistor : resistor_translator,
    elm.Conductance : conductance_translator,
    elm.VoltageSource : dc_voltage_source_translator,
    elm.CurrentSource : dc_current_source_translator,
    elm.ACVoltageSource : ac_voltage_source_translator,
    elm.ACCurrentSource : ac_current_source_translator,
    elm.RectVoltageSource : rect_voltage_source_translator,
    elm.RectCurrentSource : rect_current_source_translator,
    elm.Capacitor : capacitor_translator,
    elm.Inductance : inductance_translator,
    elm.Ground : ground_translator,
    elm.Line: none_translator,
    elm.Node: none_translator,
    elm.LabelNode: none_translator,
    elm.RealCurrentSource: linear_current_source_translator,
    elm.RealVoltageSource: linear_voltage_source_translator,
    elm.Switch: switch_translator,
    elm.VoltageLabel: none_translator,
    elm.CurrentLabel: none_translator,
    elm.PowerLabel: none_translator,
}



import numpy as np
from .network import Network
from .elements import is_ideal_current_source, is_ideal_voltage_source, is_current_source, is_voltage_source
from .supernodes import SuperNodes
from . import labelmapper as map
from . import transformers as trf
from .solution import NetworkSolution
import itertools

class DimensionError(Exception):
    ...

def admittance_connected_to(network: Network, node: str) -> complex:
    return sum(b.element.Y for b in network.branches_connected_to(node) if np.isfinite(b.element.Y))

def admittance_between(network: Network, node1: str, node2: str) -> complex:
    return sum([b.element.Y for b in network.branches_between(node1, node2) if np.isfinite(b.element.Y)])

def connected_nodes(network: Network, node: str) -> list[str]:
    return [b.node1 if b.node1 != node else b.node2 for b in network.branches_connected_to(node)]

def create_node_matrix_from_network(network: Network, node_index_mapper: map.NodeIndexMapper = map.default) -> np.ndarray:
    def node_matrix_element(i_label:str, j_label:str) -> complex:
        if i_label == j_label:
            return admittance_connected_to(passive_net, i_label)
        return -admittance_between(passive_net, i_label, j_label)
    passive_net = trf.passive_network(network)
    node_mapping = node_index_mapper(network)
    A = np.zeros((len(node_mapping), len(node_mapping)), dtype=complex)
    for (i_label, i), (j_label, j) in itertools.product(node_mapping.items(), repeat=2):
        A[i, j] = node_matrix_element(i_label, j_label)
    return A

def create_current_vector_from_network(network: Network, node_index_mapper: map.NodeIndexMapper = map.default) -> np.ndarray:
    super_nodes = SuperNodes(network)
    node_mapping = node_index_mapper(network)
    reference_node_mapping = {n: i for n, i in node_mapping.items() if super_nodes.is_reference(n)}
    b = np.zeros(len(node_mapping), dtype=complex)
    def current_sources(node: str) -> complex:
        current_sources = [b for b in network.branches_connected_to(node) if is_current_source(b.element)]
        return sum([-cs.element.I if cs.node1 == node else cs.element.I for cs in current_sources])
    def currents_of_connected_active_nodes(node: str) -> complex:
        connected_active_nodes = {n for n in connected_nodes(network, node) if super_nodes.is_active(n)}
        connected_active_nodes = {n for n in connected_active_nodes if not super_nodes.belong_to_same(n, node)}
        return sum([super_nodes.voltage_to_next_reference(cn)*admittance_between(network, cn, node) for cn in connected_active_nodes])
    def currents_of_belonging_voltage_sources(reference_node: str) -> complex:
        an = super_nodes.get_active_node(reference_node)
        connected_admittances_without_parallel_one = admittance_connected_to(network, an)-admittance_between(network, an, reference_node)
        return -super_nodes.voltage_to_next_reference(an)*connected_admittances_without_parallel_one
    for i_label, i in node_mapping.items():
        b[i] = current_sources(i_label)
        b[i] += currents_of_connected_active_nodes(i_label)
    for ref_label, i in reference_node_mapping.items():
        b[i] += current_sources(super_nodes.get_active_node(ref_label))
        b[i] += currents_of_connected_active_nodes(super_nodes.get_active_node(ref_label))
        b[i] += currents_of_belonging_voltage_sources(ref_label)
    return b
        
def open_circuit_impedance(network: Network, node1: str, node2: str, node_index_mapper: map.NodeIndexMapper = map.default) -> complex:
    if node1 == node2:
        return 0
    if any([is_ideal_voltage_source(b.element) for b in network.branches_between(node1, node2)]):
        return 0
    if network.is_zero_node(node1):
        node1, node2 = node2, node1
    network = trf.switch_ground_node(network=network, new_ground=node2)
    network = trf.passive_network(network)
    Y = create_node_matrix_from_network(network, node_index_mapper=node_index_mapper)
    Z = np.linalg.inv(Y)
    i1 = node_index_mapper(network)[node1]
    return Z[i1][i1]

def element_impedance(network: Network, element: str, node_index_mapper: map.NodeIndexMapper = map.default) -> complex:
    return open_circuit_impedance(
        network=trf.remove_element(network, element),
        node1=network[element].node1,
        node2=network[element].node2,
        node_index_mapper=node_index_mapper
    )

def open_circuit_voltage(network: Network, node1: str, node2: str) -> complex:
    if node1 == node2:
        return 0
    solution = NodalAnalysisSolution(network)
    phi1 = solution.get_potential(node=node1)
    phi2 = solution.get_potential(node=node2)
    return phi1-phi2

def short_circuit_current(network: Network, node1: str, node2: str) -> complex:
    Z = open_circuit_impedance(network, node1, node2)
    V = open_circuit_voltage(network, node1, node2)
    return V/Z

def calculate_node_voltages(Y : np.ndarray, I : np.ndarray) -> np.ndarray:
    if np.any(np.logical_not(np.isfinite(Y))):
        raise ValueError
    if np.any(np.logical_not(np.isfinite(I))):
        raise ValueError
    if Y.ndim != 2:
        raise DimensionError('dim error')
    if I.ndim != 1:
        raise DimensionError('dim error')
    m, n = Y.shape
    if n != m:
        raise DimensionError('dim error')
    return np.linalg.solve(Y, I)

class NodalAnalysisSolution:
    def __init__(self, network : Network, node_mapper: map.NodeIndexMapper = map.default) -> None:
        Y = create_node_matrix_from_network(network, node_index_mapper=node_mapper)
        I = create_current_vector_from_network(network, node_index_mapper=node_mapper)
        self._network = network
        self._super_nodes = SuperNodes(network)
        self._node_mapping = node_mapper(network)
        try:
            self._solution_vector = calculate_node_voltages(Y, I)
        except np.linalg.LinAlgError:
            self._solution_vector = np.zeros(np.size(I))
        if np.any(np.isnan(self._solution_vector)):
            self._solution_vector = np.zeros(np.size(I))

    def _select_active_node(self, branch_id: str) -> str:
        branch = self._network[branch_id]
        if self._super_nodes.is_active(branch.node1):
            return branch.node1
        return branch.node2

    def get_potential(self, node: str) -> complex:
        V_active = 0+0j
        if self._super_nodes.is_active(node):
            V_active = self._super_nodes.voltage_to_next_reference(node)
            node = self._super_nodes.next_reference(node)
        if self._network.is_zero_node(node):
            return V_active
        return self._solution_vector[self._node_mapping[node]] + V_active
    
    def get_voltage(self, branch_id: str) -> complex:
        phi1 = self.get_potential(self._network[branch_id].node1)
        phi2 = self.get_potential(self._network[branch_id].node2)
        return phi1-phi2

    def get_current(self, branch_id: str) -> complex:
        branch = self._network[branch_id]
        if is_ideal_current_source(branch.element):
            return branch.element.I
        if is_ideal_voltage_source(branch.element):
            Z = element_impedance(self._network, branch_id)
            I_branch_element = -branch.element.V/Z
            I_other_elements = short_circuit_current(
                trf.remove_element(self._network, branch_id),
                branch.node1,
                branch.node2)
            return  I_branch_element+I_other_elements
        if is_voltage_source(branch.element):
            return -(self.get_voltage(branch_id)+branch.element.V)/branch.element.Z
        return self.get_voltage(branch_id)/branch.element.Z

    def get_power(self, branch_id: str) -> complex:
        return self.get_voltage(branch_id)*self.get_current(branch_id).conjugate()

def nodal_analysis_solver(network: Network) -> NetworkSolution:
    return NodalAnalysisSolution(network)



import numpy as np
from .circuit import Circuit, transform_circuit
from ..Network import NodalAnalysis as ntw_imp

def open_circuit_impedance(circuit: Circuit, node1: str, node2: str, w: np.ndarray = np.array([0])) -> np.ndarray:
    return np.array([ntw_imp.open_circuit_impedance(transform_circuit(circuit, w0), node1, node2) for w0 in w])

def element_impedance(circuit: Circuit, element_id: str, w: np.ndarray = np.array([0])) -> np.ndarray:
    return np.array([ntw_imp.element_impedance(transform_circuit(circuit, w0), element_id) for w0 in w])

def open_circuit_dc_resistance(circuit: Circuit, node1: str, node2: str) -> float:
    return open_circuit_impedance(circuit, node1, node2, w=np.array([0]))[0].real

def element_dc_resistance(circuit: Circuit, element_id: str) -> float:
    return element_impedance(circuit, element_id, w=np.array([0]))[0].real



from typing import Protocol, Type
from dataclasses import dataclass, field
from .types import TimeDomainFunction
import numpy as np

class TransformationError(Exception):
    ...

@dataclass
class PeriodicFunction(Protocol):
    period: float
    amplitude: float
    phase: float

    @property
    def time_function(self) -> TimeDomainFunction:
        ...

@dataclass
class HarmonicCoefficients(Protocol):
    amplitude0: float
    phase0: float

    def amplitude(self, n: int) -> float:
        ...

    def phase(self, n: int) -> float:
        ...

@dataclass
class CosFunction:
    period: float
    amplitude: float
    phase: float = field(default=0)

    @property
    def time_function(self) -> TimeDomainFunction:
        return lambda t: self.amplitude*np.cos(2*np.pi/self.period*t + self.phase)

@dataclass
class CosFunctionHarmonics:
    amplitude0: float
    phase0: float

    def amplitude(self, n: int) -> float:
        if n < 0:
            raise ValueError('Fourier index must be positive.')
        if n == 1:
            return self.amplitude0
        return 0

    def phase(self, n: int) -> float:
        if n < 0:
            raise ValueError('Fourier index must be positive.')
        if n == 1:
            return self.phase0
        return 0

@dataclass
class RectFunction:
    period: float
    amplitude: float
    phase: float

    @property
    def time_function(self) -> TimeDomainFunction:
        t0 = self.phase/2/np.pi*self.period
        return np.vectorize(lambda t: self.amplitude if (t+t0) % self.period < self.period/2 else -self.amplitude)

@dataclass
class RectFunctionHarmonics:
    amplitude0: float
    phase0: float

    def amplitude(self, n: int) -> float:
        if n < 0:
            raise ValueError('Fourier index must be positive.')
        if n%2 == 0:
            return 0
        return 4/n/np.pi*self.amplitude0

    def phase(self, n: int) -> float:
        if n < 0:
            raise ValueError('Fourier index must be positive.')
        if n%2 == 0:
            return 0
        return n*self.phase0-np.pi/2

fourier_series_mapping: dict[Type[PeriodicFunction], Type[HarmonicCoefficients]] = {
    CosFunction: CosFunctionHarmonics,
    RectFunction: RectFunctionHarmonics
}

def fourier_series(time_function: PeriodicFunction) -> HarmonicCoefficients:
    try:
        return fourier_series_mapping[type(time_function)](amplitude0=time_function.amplitude, phase0=time_function.phase)
    except KeyError:
        raise TransformationError(f'No fourier coefficents found for time function of type {type(time_function).__name__}')


from typing import TypeVar, Callable, Any, Type
import schemdraw.elements

SchemdrawElement = TypeVar('SchemdrawElement', bound=schemdraw.elements.Element)
SchemdrawElementTranslator = Callable[[SchemdrawElement, tuple[str, ...]], Any]

ElementTranslatorMap = dict[Type[schemdraw.elements.Element], SchemdrawElementTranslator]